from qitip.objects import EntropicSpace


def test_from_no_random_variable() -> None:
    entropic_space = EntropicSpace(n=0)

    assert entropic_space._all_pairs == ()


def test_from_single_element() -> None:
    entropic_space = EntropicSpace(n=1)

    assert entropic_space._all_pairs == ({1},)


def test_from_two_element() -> None:
    entropic_space = EntropicSpace(n=2)

    assert entropic_space._all_pairs == (
        {1},
        {2},
        {1, 2},
    )


def test_from_three_element() -> None:
    entropic_space = EntropicSpace(n=3)

    assert entropic_space._all_pairs == (
        {1},
        {2},
        {3},
        {1, 2},
        {1, 3},
        {2, 3},
        {1, 2, 3},
    )


def test_from_four_element() -> None:
    entropic_space = EntropicSpace(n=4)

    assert len(entropic_space._all_pairs) == 2**4 - 1


def test_from_ten_element() -> None:
    entropic_space = EntropicSpace(n=10)

    assert len(entropic_space._all_pairs) == 2**10 - 1



from dataclasses import InitVar, dataclass, field
from typing import Iterable

import numpy as np
from numpy.typing import ArrayLike

from qitip.utils.converters import create_matrix_with_coefficient_list
from qitip.utils.validators import validate_matrix


def update_constraints(
    curr: np.ndarray, new: ArrayLike
) -> np.ndarray[np.float64, np.dtype[np.float64 | np.int64]]:
    # Add the new constraints into the current constraints
    # Also remove duplicate constraints

    # validate if "new" holds valid constraints
    temp = validate_matrix(m=new, dim=curr.shape[1]).reshape((-1, curr.shape[1]))

    return np.unique(np.concatenate((curr, temp)), axis=0)


@dataclass
class Constraints:
    vector_entry: dict[frozenset[int], int]
    c: InitVar[ArrayLike | None] = field(default=None)
    coefficients: np.ndarray = field(init=False)

    def __post_init__(self, c: ArrayLike | None = None):
        if (c is None) or (np.array(c) == 0).all():
            self.coefficients = np.empty((0, len(self.vector_entry)))
            return

        self.coefficients = update_constraints(
            curr=np.empty((0, len(self.vector_entry))), new=c
        )


class ConstraintsBuilder:
    # Constraints builder is chaacterized by the vector entry
    def __init__(self, vector_entry: dict[frozenset[int], int]) -> None:
        self._vector_entry: dict[frozenset[int], int] = vector_entry

    def __call__(self, c: ArrayLike | None = None) -> Constraints:
        """
        When we use cb = ConstraintsBuilder(vectro_entry)
        cb(c) should give us a new Constraints object

        Args:
            c (ArrayLike | None, optional): Constraints in the form
             of a 2D matrix. Defaults to None.
        """

        return Constraints(vector_entry=self._vector_entry, c=c)

    def from_coefficients(
        self, c: Iterable[dict[Iterable[int] | int, float]]
    ) -> Constraints:
        """
        As the number of coefficients increases, it is easier to
        create an object by specifying the coefficients

        Args:
            c (Iterable[dict[Iterable[int]  |  int, float]]):
            A collection of dictionaries where each dictionary defines
            the coefficients of marginal entropies

            format: {marginal system: coefficients}
        """
        return Constraints(
            vector_entry=self._vector_entry,
            c=create_matrix_with_coefficient_list(vector_entry=self._vector_entry)(c),
        )



from qitip.objects import Constraints

vector_entry: dict[frozenset[int], int] = {
    frozenset((1,)): 0,
    frozenset((2,)): 1,
    frozenset((1, 2)): 2,
}


def test_empty_constraint_initialization():
    empty_constraint = Constraints(vector_entry=vector_entry)
    assert empty_constraint.coefficients.shape == (0, len(vector_entry))


def test_single_constraint_initialization_with_vector():
    single_constraint = Constraints(vector_entry, (1, 0, 0))
    assert single_constraint.coefficients.shape == (1, len(vector_entry))


def test_single_constraint_initialization_with_matrix():
    single_constraint = Constraints(vector_entry, ((1, 0, 0),))
    assert single_constraint.coefficients.shape == (1, len(vector_entry))


def test_multiple_constraints_initialization():
    num_constraints: int = 2
    c = [
        [(i + len(vector_entry) * _) for i in range(len(vector_entry))]
        for _ in range(num_constraints)
    ]
    multi_constraints = Constraints(vector_entry, c)
    assert multi_constraints.coefficients.shape == (num_constraints, len(vector_entry))



import numpy as np
from numpy.typing import NDArray
from qitip.objects import EntropicSpace
from qitip.prover import Prover

# from qitip.quantum_inequalities import QuantumElementalInequalities

"""
The prover can also be applied in the classical regime by replacing 
QuantumElementalInequalities(space).ELEMENtAL with ShannonInequality(space).ELEMMENTAL
"""


def test_quantum_strong_subadditivity() -> None:
    n: int = 3
    space = EntropicSpace(n)
    # prover = Prover(
    #     elemental=QuantumElementalInequalities(
    #         vector_entry=space.vector_entry
    #     ).get_elementals(),
    #     n=n,
    # )
    prover = Prover(space=space)
    inequality: NDArray[np.float64] = np.zeros(len(space.vector_entry))
    inequality[space.vector_entry[frozenset({1, 2, 3})]] = -1
    inequality[space.vector_entry[frozenset({3})]] = -1
    inequality[space.vector_entry[frozenset({1, 3})]] = 1
    inequality[space.vector_entry[frozenset({2, 3})]] = 1

    assert prover._check_type(
        inequality=inequality, constraints=np.empty((0, len(space.vector_entry)))
    )


def test_quantum_weak_monotonacity() -> None:
    n: int = 3
    space = EntropicSpace(n=n)
    # prover = Prover(
    #     elemental=QuantumElementalInequalities(
    #         vector_entry=space.vector_entry
    #     ).get_elementals(),
    #     n=n,
    # )
    prover = Prover(space=space)

    inequality = np.zeros(len(space.vector_entry))

    inequality[space.vector_entry[frozenset({1})]] = -1
    inequality[space.vector_entry[frozenset({2})]] = -1
    inequality[space.vector_entry[frozenset({1, 3})]] = 1
    inequality[space.vector_entry[frozenset({2, 3})]] = 1

    assert prover._check_type(
        inequality=inequality, constraints=np.empty((0, len(space.vector_entry)))
    )


def test_conditional_von_neumann_entropy() -> None:
    n: int = 2

    space = EntropicSpace(n=n)
    # prover = Prover(
    #     elemental=QuantumElementalInequalities(
    #         vector_entry=space.vector_entry
    #     ).get_elementals(),
    #     n=n,
    # )
    prover = Prover(space=space)
    inequality = np.zeros(len(space.vector_entry))

    inequality[space.vector_entry[frozenset({1, 2})]] = 1
    inequality[space.vector_entry[frozenset({2})]] = -1

    assert (
        prover._check_type(
            inequality=inequality,
            constraints=np.empty((0, len(space.vector_entry))),
        )
        is False
    )


def test_constrained_non_von_neumann_type() -> None:
    n: int = 4

    space = EntropicSpace(n=n)
    # prover = Prover(
    #     elemental=QuantumElementalInequalities(
    #         vector_entry=space.vector_entry
    #     ).get_elementals(),
    #     n=n,
    # )
    prover = Prover(space=space)

    constraints = np.zeros((3, len(space.vector_entry)))
    # I(C;A|B) = 0
    constraints[0][space.vector_entry[frozenset({2, 3})]] = 1
    constraints[0][space.vector_entry[frozenset({1, 2})]] = 1
    constraints[0][space.vector_entry[frozenset({1, 2, 3})]] = -1
    constraints[0][space.vector_entry[frozenset({2})]] = -1
    # I(B;C|A) = 0
    constraints[1][space.vector_entry[frozenset({1, 2})]] = 1
    constraints[1][space.vector_entry[frozenset({1, 3})]] = 1
    constraints[1][space.vector_entry[frozenset({1, 2, 3})]] = -1
    constraints[1][space.vector_entry[frozenset({1})]] = -1
    # I(A;B|D) = 0
    constraints[2][space.vector_entry[frozenset({1, 4})]] = 1
    constraints[2][space.vector_entry[frozenset({2, 4})]] = 1
    constraints[2][space.vector_entry[frozenset({1, 2, 4})]] = -1
    constraints[2][space.vector_entry[frozenset({4})]] = -1

    inequality = np.zeros(len(space.vector_entry))
    # I(C;D) >= I(C;AB)
    # equivalently, (S_{c} + S_{d} - S_{c,d}) - (S_{c} + S_{ab} - S_{abc}) >= 0
    inequality[space.vector_entry[frozenset({4})]] = 1
    inequality[space.vector_entry[frozenset({3, 4})]] = -1
    inequality[space.vector_entry[frozenset({1, 2})]] = -1
    inequality[space.vector_entry[frozenset({1, 2, 3})]] = 1

    assert prover._check_type(inequality=inequality, constraints=constraints) is False


def test_quantum_zhang_yeung_inequality() -> None:
    n: int = 4
    space = EntropicSpace(n)

    prover = Prover(space=space)

    inequality = np.zeros(len(space.vector_entry))
    # Zhang-Yeung Inequality in canonical form
    inequality[space.vector_entry[frozenset({1})]] = -2
    inequality[space.vector_entry[frozenset({2})]] = -2
    inequality[space.vector_entry[frozenset({3})]] = -1
    inequality[space.vector_entry[frozenset({1, 4})]] = 1
    inequality[space.vector_entry[frozenset({2, 4})]] = 1
    inequality[space.vector_entry[frozenset({3, 4})]] = -1
    inequality[space.vector_entry[frozenset({1, 2})]] = 3
    inequality[space.vector_entry[frozenset({1, 3})]] = 3
    inequality[space.vector_entry[frozenset({2, 3})]] = 3
    inequality[space.vector_entry[frozenset({1, 2, 3})]] = -4
    inequality[space.vector_entry[frozenset({1, 2, 4})]] = -1

    assert (
        prover._check_type(
            inequality, constraints=np.empty((0, len(space.vector_entry)))
        )
        is False
    )



from itertools import combinations

import numpy as np
from numpy.typing import NDArray


class QuantumElementalInequalities:
    """
    The elemental inequalities are based on the paper by Pippenger

    1. Delta(I, J) >= 0 which are referred to as type 1
        I - J = {i} and J - I = {j} and i >= j
    2. E(I, J) >=0 which are referred to as type 2
        I ∩ J = {k} and k+1 ∈ I and I ∪ J = {1,2,...,N}
    """

    def __init__(self, vector_entry: dict[frozenset[int], int]):
        self.vector_entry: dict[frozenset[int], int] = vector_entry
        self.entire_system: frozenset[int] = max(self.vector_entry.keys())
        self.n: int = max(self.entire_system)

        if self.n < 2:
            raise ValueError(
                f"Number of quantum systems should be >= 2; {self.n} is given instead."
            )

    def _get_type_1_elemental_vector(self, i: int, j: int) -> NDArray[np.int64]:
        remaining: frozenset[int] = self.entire_system - {i, j}

        elemental_i_j: NDArray = np.empty((0, len(self.vector_entry)))

        # Combinations run from r = 0 to r = num of all remainings
        for r in range(len(remaining) + 1):
            for intersection in combinations(remaining, r):
                set_i = frozenset(intersection).union({i})
                set_j = frozenset(intersection).union({j})

                vector = np.zeros(len(self.vector_entry))
                vector[self.vector_entry[set_i]] = 1
                vector[self.vector_entry[set_j]] = 1
                vector[self.vector_entry[frozenset(set_i.union(set_j))]] = -1

                # Exclude intersections being empty
                if intersection != tuple():
                    vector[self.vector_entry[frozenset(intersection)]] = -1

                elemental_i_j = np.vstack((elemental_i_j, vector))

        return elemental_i_j

    def _get_type_2_elemental_vector(self, k: int) -> NDArray[np.int64]:
        intersection = {k}

        i_diff_j = {(k + 1) % self.n if k + 1 > self.n else k + 1}

        remaining = self.entire_system - intersection - i_diff_j

        elemental_k: NDArray[np.int64] = np.empty(
            (0, len(self.vector_entry)), dtype=np.int8
        )

        for r in range(0, len(remaining) + 1):
            for i_remaning in combinations(remaining, r):
                vec = np.zeros(len(self.vector_entry))

                i_remaning = frozenset(i_remaning)
                j_remaining = remaining.difference(i_remaning)
                set_i = (i_remaning.union(i_diff_j)).union(intersection)
                set_j = j_remaining.union(intersection)

                vec[self.vector_entry[set_i]] = 1
                vec[self.vector_entry[set_j]] = 1
                vec[self.vector_entry[set_i.difference(set_j)]] = -1
                if j_remaining != frozenset():
                    vec[self.vector_entry[set_j.difference(set_i)]] = -1

                elemental_k: NDArray[np.int64] = np.vstack((elemental_k, vec))

        return elemental_k

    def _get_all_type_1(self):
        all_type_1: NDArray = np.empty((0, len(self.vector_entry)))
        for group in self.vector_entry.keys():
            if len(group) > 2:
                break
            elif len(group) == 2:
                i, j = group
                all_type_1 = np.vstack(
                    (all_type_1, self._get_type_1_elemental_vector(i, j))
                )
        return all_type_1

    def _get_all_type_2(self):
        all_type_2 = np.empty((0, len(self.vector_entry)))
        for k in range(1, self.n + 1):
            all_type_2 = np.vstack((all_type_2, self._get_type_2_elemental_vector(k=k)))

        return all_type_2

    def get_elementals(self) -> NDArray[np.float64]:
        return np.vstack((self._get_all_type_1(), self._get_all_type_2()))



from typing import Optional

from qitip.objects import (
    Constraints,
    ConstraintsBuilder,
    EntropicSpace,
    Inequality,
    InequalityBuilder,
    InfoType,
    SpacePool,
    TypeResult,
    result_director,
)
from qitip.prover import Prover, ProverPool
from qitip.utils.converters import CoefficientsToDict


class Qitip:
    _space_pool: SpacePool = SpacePool()
    _prover_pool: ProverPool = ProverPool()

    def __init__(self, n: int):
        self._space: EntropicSpace = self._space_pool.get(n)
        self._prover: Prover = self._prover_pool.get(space=self._space)
        self.inequality: InequalityBuilder = InequalityBuilder(
            vector_entry=self._space.vector_entry
        )
        self.constraints: ConstraintsBuilder = ConstraintsBuilder(
            vector_entry=self._space.vector_entry
        )

    def embed(self, obj: InfoType) -> InfoType:
        """_summary_
        Embed the existing InfoType (can be either Inequality or Constraints) in
        higher dimensinoal space

        Args:
            obj (InfoType): Inequality or Constraints type

        Returns:
            InfoType: Type has to agree with that of the obj type
        """
        if not isinstance(obj, Inequality | Constraints):
            raise TypeError("Only Inequality and Constraints can be embedded ...")

        if len(obj.vector_entry) > len(self._space.vector_entry):
            raise ValueError("Can only embed an object to higher dimensions!")

        elif len(obj.vector_entry) == len(self._space.vector_entry):
            return obj

        # If the dim(obj) < dim(Qitip), actual embedding
        elif isinstance(obj, Inequality):
            return self.inequality.from_coefficients(
                CoefficientsToDict.convert_vector(obj.vector_entry, obj.coefficients)
            )

        else:
            return self.constraints.from_coefficients(
                CoefficientsToDict.convert_matrix(obj.vector_entry, obj.coefficients)
            )

    def is_vn_type(
        self, inequality: Inequality, constraints: Optional[Constraints] = None
    ) -> TypeResult:
        return result_director(
            prover=self._prover, inequality=inequality, constraints=constraints
        )

    def check_vn_result(
        self, inequality: Inequality, constraints: Optional[Constraints] = None
    ):
        print(self.is_vn_type(inequality, constraints).message)

    @property
    def vector_entry(self):
        return self._space.vector_entry



from itertools import compress
from typing import Any, Callable, Iterable

from numpy import dtype, float64, int8, ndarray


def convert_iterable_int_to_set(key: Iterable[int] | int) -> frozenset[int]:
    if isinstance(key, int):
        return frozenset({key})
    elif all([isinstance(i, int) for i in key]):
        return frozenset(key)
    else:
        raise TypeError(
            f"Key, {key} can either be an integer or a sequence of integers."
        )


def create_vector_with_coefficient(
    vector_entry: dict[frozenset[int], int]
) -> Callable[[dict[Iterable[int] | int, float]], tuple[float, ...]]:
    def assign_coefficients(
        coefficients: dict[Iterable[int] | int, float]
    ) -> tuple[float, ...]:
        sys_coefficients: dict[frozenset[int], float] = {
            convert_iterable_int_to_set(k): v for k, v in coefficients.items()
        }

        # Create a tuple of coeffcients in the order of vector entry
        # If the coefficient is not assigned by the user, it is set to 0
        # Otherwise, use the user-assigned coefficient
        return tuple(
            sys_coefficients[key] if sys_coefficients.get(key, None) else 0
            for key in vector_entry.keys()
        )

    return assign_coefficients


def create_matrix_with_coefficient_list(
    vector_entry: dict[frozenset[int], int]
) -> Callable[
    [Iterable[dict[Iterable[int] | int, float]]], tuple[tuple[float, ...], ...]
]:
    def assign_coefficients(
        coefficient_list: Iterable[dict[Iterable[int] | int, float]]
    ) -> tuple[tuple[float, ...], ...]:
        return tuple(
            [
                create_vector_with_coefficient(vector_entry)(coefficients)
                for coefficients in coefficient_list
            ]
        )

    return assign_coefficients


def vector_entry_to_ordered_sys(
    vector_entry: dict[frozenset[int], int]
) -> tuple[tuple[int, ...], ...]:
    return tuple(
        tuple(k) for k, _ in sorted(vector_entry.items(), key=lambda item: item[1])
    )


def frozenset_to_marginal_entropy(parties: frozenset[int]) -> str:
    return f"S({', '.join(tuple(str(party) for party in parties))})"


def canonical_to_expression(party_coefficient: dict[frozenset[int], float]) -> str:
    exp: str = " ".join(
        [
            (
                f"+ {coefficient} * {frozenset_to_marginal_entropy(party)}"
                if coefficient >= 0
                else f"- {abs(coefficient)} * {frozenset_to_marginal_entropy(party)}"
            )
            for party, coefficient in party_coefficient.items()
        ]
    )

    return exp if exp[0] == "-" else exp[2:]


class CoefficientsToDict:
    # This function takes an object (can either be Inequality or Constraints) to
    # 1D arraylike or 2D arraylike
    @staticmethod
    def convert_vector(
        vector_entry: dict[Any, int],
        coefficients: ndarray[float64, dtype[float64]],
    ) -> dict[Any, float]:
        non_zeros: ndarray[ndarray[bool, dtype[int8]], dtype[int8]] = coefficients != 0

        # Ensures the sys is in the same order as the vectors
        sys_in_order: tuple[tuple[int, ...], ...] = vector_entry_to_ordered_sys(
            vector_entry
        )

        # since each Inequality only contains one inequality
        return dict(
            zip(
                compress(sys_in_order, non_zeros),
                compress(coefficients, non_zeros),
            )
        )

    @staticmethod
    def convert_matrix(
        vector_entry: dict[Any, int],
        coefficients: ndarray[float64, dtype[float64]],
    ) -> list[dict[Any, float]]:
        non_zeros: ndarray[ndarray[bool, dtype[int8]], dtype[int8]] = coefficients != 0

        # Ensures the sys is in the same order as the vectors
        sys_in_order: tuple[tuple[int, ...], ...] = vector_entry_to_ordered_sys(
            vector_entry
        )

        return [
            dict(
                zip(
                    compress(sys_in_order, non_zero),
                    compress(coefficients[index], non_zero),
                )
            )
            for index, non_zero in enumerate(non_zeros)
        ]



from dataclasses import dataclass, field
from typing import Optional

from numpy import array

from qitip.objects import Constraints, Inequality
from qitip.prover import Prover
from qitip.utils.converters import CoefficientsToDict, canonical_to_expression


@dataclass
class TypeResult:
    status: Optional[bool] = field(init=False, default=None)
    message: str = field(init=False, default="")


class ResultBuilder:
    def __init__(self, prover: Prover) -> None:
        self._prover: Prover = prover
        self.reset()

    @property
    def result(self) -> TypeResult:
        return self._result

    def reset(self) -> None:
        self._result: TypeResult = TypeResult()

    def process_type(self, inequality: Inequality, constraints: Constraints) -> None:
        self._result.status = self._prover._check_type(
            inequality.coefficients, constraints.coefficients
        )

        # Abbreviation for von-Neumann type
        vn_message: str = "It's von-Neumann type inequality.\n\nIt can be proved by summing up the following:\n"
        # Abbreviation for non-Provable type
        np_message: str = (
            "Not provable by Quantum ITIP:(\n\nOne can try to disprove by using:\n"
        )
        self._result.message += vn_message if self._result.status else np_message

    def process_used_inequality_constraints(
        self, inequality: Inequality, constraints: Constraints
    ):
        if self._result.status is None:
            raise NotImplementedError(
                f"Check type before calling {self.process_used_inequality_constraints.__qualname__}"
            )

        # Mapping from elemental inequalities to vector entries of used _inequalities
        inequality_entry: dict[tuple[int, ...], int] = {
            tuple(elemental): index
            for index, elemental in enumerate(self._prover.elemental)
        }

        constraints_entry: dict[tuple[float, ...], int] = {
            tuple(constraint): index
            for index, constraint in enumerate(constraints.coefficients)
        }

        # the inequality is von-Neumann type
        if self._result.status:
            used_inequalities, used_constraints = self._prover._shortest_proof(
                inequality=inequality.coefficients, constraints=constraints.coefficients
            )

            for elemental, coefficient in CoefficientsToDict.convert_vector(
                vector_entry=inequality_entry, coefficients=used_inequalities
            ).items():
                # Notice that the keys of the inequality_entries are "tuples" !!!!!
                self._result.message += f"{coefficient} * [{canonical_to_expression(CoefficientsToDict.convert_vector(vector_entry=inequality.vector_entry, coefficients=array(elemental)))}] >= 0\n"

            for constraint, coefficient in CoefficientsToDict.convert_vector(
                vector_entry=constraints_entry, coefficients=used_constraints
            ).items():
                # The negative 1 comes from the expression of duality
                self._result.message += f"{-1*coefficient} * [{canonical_to_expression(CoefficientsToDict.convert_vector(vector_entry=constraints.vector_entry, coefficients=array(constraint)))}] = 0\n"

        # not provable by quantum ITIP
        else:
            temp_used_inequalities, temp_used_constriants = (
                self._prover._shortest_counter_proof(
                    inequality=inequality.coefficients,
                    constraints=constraints.coefficients,
                )
            )

            used_inequalities = (temp_used_inequalities != 0).astype(int)
            used_constraints = (temp_used_constriants != 0).astype(int)

            for elemental, _ in CoefficientsToDict.convert_vector(
                vector_entry=inequality_entry, coefficients=used_inequalities
            ).items():
                self._result.message += f"{canonical_to_expression(CoefficientsToDict.convert_vector(vector_entry=inequality.vector_entry, coefficients=array(elemental)))} = 0\n"

            for constraint, _ in CoefficientsToDict.convert_vector(
                vector_entry=constraints_entry, coefficients=used_constraints
            ).items():
                self._result.message += f"{canonical_to_expression(CoefficientsToDict.convert_vector(vector_entry=constraints.vector_entry, coefficients=array(constraint)))} = 0"


def result_director(
    prover: Prover, inequality: Inequality, constraints: Optional[Constraints] = None
):
    if constraints is None:
        _constraints: Constraints = Constraints(vector_entry=inequality.vector_entry)
    else:
        _constraints = constraints

    builder: ResultBuilder = ResultBuilder(prover=prover)
    builder.process_type(inequality=inequality, constraints=_constraints)
    builder.process_used_inequality_constraints(
        inequality=inequality, constraints=_constraints
    )
    return builder.result



from qitip.objects.constraints import Constraints, ConstraintsBuilder
from qitip.objects.entrospace import EntropicSpace, SpacePool
from qitip.objects.inequality import Inequality, InequalityBuilder
from qitip.objects.type_result import TypeResult, result_director
from qitip.objects.typings import InfoType



from typing import Iterable, Sequence

import numpy as np
from qitip.objects import (
    Constraints,
    ConstraintsBuilder,
    Inequality,
    InequalityBuilder,
)

# Test with bipartite system for simplicity
vector_entry = {frozenset({1}): 0, frozenset({2}): 1, frozenset({1, 2}): 2}


def test_call_inequality_builder() -> None:
    builder: InequalityBuilder = InequalityBuilder(vector_entry=vector_entry)

    vec: Sequence[float] = (0, 0, 1)
    inequality: Inequality = builder((0, 0, 1))

    assert inequality.coefficients.shape == (len(vector_entry),)
    assert (inequality.coefficients == np.array(vec)).all()


def test_inequality_built_from_coefficients() -> None:
    builder: InequalityBuilder = InequalityBuilder(vector_entry)

    vec: dict[Iterable[int] | int, float] = {(1, 2): 1, (1): -1}

    inequality: Inequality = builder.from_coefficients(vec)

    assert inequality.coefficients.shape == (len(vector_entry),)
    assert (inequality.coefficients == np.array((-1, 0, 1))).all()


def test_constraint_built_from_none_input() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)

    constraints: Constraints = builder()

    assert constraints.coefficients.size == 0
    assert constraints.coefficients.shape == (0, len(vector_entry))


def test_constraint_built_from_single_vec() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)

    constraints: Constraints = builder((1, 1, -1))

    assert constraints.coefficients.shape == (1, len(vector_entry))
    assert (constraints.coefficients == np.array([[1, 1, -1]])).all()


def test_constraints_built_from_calling_with_multiple_vecs() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)

    c = ((1, -1, 0), (-1, 0, 1))

    constraints: Constraints = builder(c)

    assert constraints.coefficients.shape == (len(c), len(vector_entry))
    assert np.array_equiv(
        np.sort(constraints.coefficients, axis=0), np.sort(np.array(c), axis=0)
    )


def test_constraints_built_from_single_coefficients() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)

    c: tuple[dict[Iterable[int] | int, float]] = ({(1, 2): 1, (1): -1},)

    constraints: Constraints = builder.from_coefficients(c)

    assert constraints.coefficients.shape == (1, len(vector_entry))
    assert np.array_equiv(constraints.coefficients, np.array([[-1, 0, 1]]))


def test_constraints_built_from_multiple_coefficients() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)

    c: tuple[dict[Iterable[int] | int, float], ...] = (
        {(1, 2): 1, (1): -1},
        {(1,): -1, (2): 1},
    )

    constraints = builder.from_coefficients(c)

    assert constraints.coefficients.shape == (len(c), len(vector_entry))
    assert np.array_equiv(
        np.sort(constraints.coefficients, axis=0),
        np.sort(np.array([[-1, 0, 1], [-1, 1, 0]]), axis=0),
    )


def test_constraints_built_from_empty_coefficients() -> None:
    builder: ConstraintsBuilder = ConstraintsBuilder(vector_entry)
    c = (dict(),)

    constraints = builder.from_coefficients(c)

    assert constraints.coefficients.size == 0


def test_constraints_built_from_nothing():
    builder = ConstraintsBuilder(vector_entry)

    empty_constraints: Constraints = builder.from_coefficients([])

    assert empty_constraints.coefficients.size == 0
